package com.antlersoft.jdisassemble;

/**
 * <p>Title: jdisassemble</p>
 * <p>Description: disassembler for Java class files</p>
 * <p>Copyright (c) 2000, 2003  Michael A. MacDonald<p>
 * ----- - - -- - - --
 * <p>
 *     This package is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * <p>
 *     This package is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * <p>
 *     You should have received a copy of the GNU General Public License
 *     along with the package (see gpl.txt); if not, see www.gnu.org
 * <p>
 * ----- - - -- - - --
 * <p>Company: antlersoft</p>
 * @author Michael Macdonald
 * @version 1.0
 */

import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.PrintStream;

import java.util.ArrayList;
import java.util.Iterator;

import com.antlersoft.classwriter.*;

public class Disassemble
{
    private void writeMethod( MethodInfo method, ClassWriter w, PrintStream out)
    throws CodeCheckException
    {
        CheckedInstruction[] instructions=method.getCodeAttribute().codeCheck();
        String[] branch_dest=new String[instructions.length];
        String[] labels=new String[instructions.length];
        int  label_count=0;
        for ( int i=0; i<instructions.length; i++)
        {
            for ( Iterator j=instructions[i].previousCheckedInstructions.iterator();
                  j.hasNext();)
            {
                CheckedInstruction prev=(CheckedInstruction)j.next();
                if ( prev.instruction.getInstructionStart()+prev.instruction.getLength()!=
                     instructions[i].instruction.getInstructionStart())
                {
                    int prev_index=method.getCodeAttribute().indexAtOffset(
                        new InstructionPointer( prev.instruction.getInstructionStart()));
                    if ( labels[i] == null)
                    {
                        label_count++;
                        labels[i]="L"+label_count;
                    }
                    if ( branch_dest[prev_index]==null)
                        branch_dest[prev_index]=labels[i];
                    else
                        branch_dest[prev_index]=branch_dest[prev_index]+","+labels[i];
                }
            }
        }
        ArrayList fields=new ArrayList( w.getFields());
        out.println( method.getType()+" "+method.getName());
        for ( int i=0; i<instructions.length; i++)
        {
            String mnemonic=instructions[i].instruction.getOpCode().getMnemonic();
            out.print( instructions[i].instruction.getInstructionStart());
            out.print( '\t');
            if ( labels[i] != null)
                out.print( labels[i]);
            out.print( "\t");
            out.print( instructions[i].instruction.getOpCode().getMnemonic());
            int length=instructions[i].instruction.getLength();
            if ( length==2 || length==3 || length==5)
            {
                out.print( ' ');
                int arg=instructions[i].instruction.operandsAsInt();
                out.print( arg);
                if ( mnemonic.indexOf( "invoke")!= -1 )
                {
                    out.print( " // ");
                    ClassWriter.CPTypeRef ref=w.getTypeRef(
                        arg/
                        ( mnemonic.indexOf( "interface")== -1 ? 1 : 65536));
                    out.print( w.getClassName( ref.getClassIndex()));
                    out.print( '.');
                    out.print( ref.getSymbolName());
                    out.print( ref.getSymbolType());
                }
                else if ( mnemonic.equals( "new") || mnemonic.equals( "checkcast") || mnemonic.equals( "instanceof"))
                {
                    out.print( " // ");
                    out.print( w.getClassName( arg));
                }
                else if ( mnemonic.equals( "ldc"))
                {
                    out.print( " // \"");
                    try
                    {
                        out.print( w.getStringConstant( arg));
                        out.print( '"');
                    }
                    catch ( Exception e)
                    {
                    }
                }
                else if ( mnemonic.indexOf( "field")==3 || mnemonic.indexOf( "static")==3)
                {
                    out.print( " // ");
                    ClassWriter.CPTypeRef ref=w.getTypeRef( arg);
                    out.print( ref.getSymbolType());
                    out.print( ' ');
                    out.print( w.getClassName( ref.getClassIndex()));
                    out.print( '.');
                    out.print( ref.getSymbolName());
                  }
            }
            if ( branch_dest[i] != null)
            {
                out.print( " -- ");
                out.print( branch_dest[i]);
            }
            out.println();
        }
    }
    public void disassemble( ClassWriter w, PrintStream out)
    throws CodeCheckException
    {
        for ( Iterator i=w.getMethods().iterator(); i.hasNext();)
        {
            MethodInfo method=(MethodInfo)i.next();
            writeMethod( method, w, out);
            //break;
        }
    }

    public static void main(String[] args)
    throws Exception
    {
        ClassWriter w=new ClassWriter();
        w.readClass( new BufferedInputStream( new FileInputStream( args[0])));
        Disassemble q=new Disassemble();
        q.disassemble( w, System.out);
    }
}